#!/usr/bin/env python

import math
import sys
import gc
import rospy
import re
import array
import numpy as np
import sys, traceback
import multiprocessing
import time
import logging
import os
from datetime import datetime
from geometry_msgs.msg import Quaternion
from tf.transformations import euler_from_quaternion, quaternion_from_euler
from nav_msgs.msg import Odometry
from gazebo_msgs.srv import GetModelState
from gazebo_msgs.srv import GetWorldProperties

BotDist = float(sys.argv[1])#Erstes Argumnet
SafeDist = float(sys.argv[2])#Zweites argument
DestinationX = float(sys.argv[3])#Ziellinie
Timeout = float(sys.argv[4])#timeout
yawBot = 0.0
WriteFlag = 0

logger = logging.getLogger("EventLog")
logger.setLevel(logging.INFO)

# create a file handler
handler = logging.FileHandler('location_monitor.log', mode='w')
handler.setLevel(logging.INFO)

# create a logging format
formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)

# add the file handler to the logger
logger.addHandler(handler)


# y
# ^
# |
# |  -------          Box muss im rechtem Winkel zur x-Achse sein!
# |  |     |
# |  |     |
# |  -------
# |
# ---------------->x


class Block:
    def __init__(self, name, relative_entity_name, radius):
        self._name = name
        self._relative_entity_name = relative_entity_name
        self._radius = radius


class Detection:
    _blockListDict = {}
    def show_gazebo_models(self, botx, boty):
     global Timeout
     global DestinationX
     global WriteFlag
     global logger
     get_world_properties = rospy.ServiceProxy('/gazebo/get_world_properties', GetWorldProperties)
     resp1 = get_world_properties()
     if resp1.sim_time >= Timeout:
         rospy.loginfo("Simulation Timeout!")
         if WriteFlag == 0:
            WriteFlag = 1
            logger.info("[T:" + str(resp1.sim_time) + "] " + "Simulation Timeout!")
            os._exit(1)
     if WriteFlag == 0:
         logger.info("[T:" + str(resp1.sim_time) + "] " + "Bot X: " + str(botx))
         logger.info("[T:" + str(resp1.sim_time) + "] " + "Bot Y: " + str(boty))
     if botx >= DestinationX and DestinationX != 0.0:
         rospy.loginfo("Bot reached Destination!")
         if WriteFlag == 0:
            WriteFlag = 1
            logger.info("[T:" + str(resp1.sim_time) + "] " + "Bot reached Destination!")
            os._exit(1)
     if botx > 0 and boty > 0:
        i = 0
        for model in resp1.model_names:
            if ("ball" in str(model)) or ("cylinder" in str(model)) or ("box" in str(model) or ("sphere" in str(model))):
                if("R_" in str(model)):
                    keyword = "R_"
                    search = str(model)
                    afsearch = search.partition(keyword)
                    matches = re.findall(r"[-+]?\d*\.\d+|\d+", afsearch[2])
                else:
                    keyword = "S_"
                    search = str(model)
                    afsearch = search.partition(keyword)
                    matches = re.findall(r"[-+]?\d*\.\d+|\d+", afsearch[2])
                rad = 0
                if matches:
                    if "box" in str(model):
                        rad = (matches[0])
                        rad = float(rad) / 2.0
                    else:
                        rad = (matches[0])
                s = 'block_' + str(i)
                self._blockListDict.update({s: Block(model, 'link', rad)})
                i = i + 1
        try:
            model_coordinates = rospy.ServiceProxy('/gazebo/get_model_state', GetModelState)
            for block in self._blockListDict.itervalues():
                blockName = str(block._name)
                if blockName in resp1.model_names:
                    resp_coordinates = model_coordinates(blockName, block._relative_entity_name)
                    l_x = resp_coordinates.pose.position.x
                    l_y = resp_coordinates.pose.position.y
                    rospy.loginfo(blockName)
                    rospy.loginfo("X: " + str(l_x))
                    rospy.loginfo("Y: " + str(l_y))
                    if WriteFlag == 0:
                        logger.info(str(blockName))
                        logger.info("[T:" + str(resp1.sim_time) + "] " + str(blockName) + " X: " + str(l_x))
                        logger.info("[T:" + str(resp1.sim_time) + "] " + str(blockName) + " Y: " + str(l_y))
                    if block._radius != 0:

                        if "box" in str(blockName):
                            box_orient_list = [resp_coordinates.pose.orientation.x,resp_coordinates.pose.orientation.y,resp_coordinates.pose.orientation.z,resp_coordinates.pose.orientation.w]
                            (roll, pitch, yaw) = euler_from_quaternion(box_orient_list)
                            yaw_degrees = yaw * 180.0 / math.pi
                            if yaw_degrees < 0:
                                yaw_degrees += 360.0
                            boxkdetection(float(block._radius), botx, boty, l_x, l_y, yaw_degrees, resp1.sim_time)
                        else: #if "sphere" , "cylinder" , "ball" etc.
                            objectWithRadiusDetection(float(block._radius), botx, boty, l_x, l_y, str(resp1.sim_time))


                    else:
                        rospy.loginfo("Dist: " + str(dist))
                        if WriteFlag == 0:
                            logger.info("[T:" + str(resp1.sim_time) + "] " + "Dist: " + str(dist))

        except rospy.ServiceException as e:
            rospy.loginfo("Get Model State service call failed:  {0}".format(e))
     else:
         rospy.loginfo("Bot out of tracking Range!((x,y)<0)")
         if WriteFlag == 0:
             logger.info("[T:" + str(resp1.sim_time) + "] " + "Bot out of tracking Range!((x,y)<0)")


def distance(x1, y1, x2, y2):
    dist = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
    return dist

def objectWithRadiusDetection(radius, botx, boty, l_x, l_y, simtime):
    global WriteFlag
    global logger
    global SafeDist
    dist = distance(botx, boty, l_x, l_y)
    safedist = radius + SafeDist  # Radius/Size + Bot front Rad + safe dist
    rospy.loginfo("Safedist: " + str(safedist))
    rospy.loginfo("Distance: " + str(dist))
    if WriteFlag == 0:
        logger.info("Safedist.: " + str(safedist))
        logger.info("[T:" + simtime + "] " + "Dist. To Collision Point: " + str(dist))
    if dist < safedist:
        rospy.loginfo("Collision!")
        if WriteFlag == 0:
            logger.info("[T:" + simtime + "] " + "Collision!")
            os._exit(1)


def boxkdetection(size, botx, boty, l_x, l_y, yaw_degrees, simtime):
    global f
    global WriteFlag
    if (size * 2.0) > 5.0:
        rospy.loginfo("Box to big, > 5m")
        if WriteFlag == 0:
            logger.info("[T:" + str(simtime) + "] " + "Box Invalid (size > 5m)")
    else:
        # rospy.loginfo("SIZEBOX" + str(size))
        safedist =  SafeDist
        #rospy.loginfo(str(safedist))
        if WriteFlag == 0:
            logger.info("Safedist.: " + str(safedist))
        xR = l_x + size * math.cos(math.radians(yaw_degrees))
        yL = l_y + size * math.sin(math.radians(yaw_degrees))


        lineA = ((botx, boty), (l_x, l_y))
        lineB = ((l_x, l_y), (xR, yL))

        angle = angle_between(lineA, lineB)
        disangle = angle
        angle1 = angle
        if angle > 45:
            disangle = 90 - angle;
        #rospy.loginfo("Angle: " + str(angle))#korrekt
        dis = (size) / math.cos(math.radians(disangle))


        if botx < l_x and boty > l_y:
            angle1 = 180 - angle
        if botx < l_x and boty < l_y:
            angle1 = 180 + angle
        if botx > l_x and boty < l_y:
            angle1 = 360 - angle

        #rospy.loginfo("Angle1: " + str(angle2))
        dynX = l_x + dis * math.cos(math.radians(angle1))
        dynY = l_y + dis * math.sin(math.radians(angle1))

        #dynX = l_x + dis * math.cos(math.radians(angle2))
        #dynY = l_y + dis * math.sin(math.radians(angle2))

        #rospy.loginfo("DynX : " + str(dynX))
        #rospy.loginfo("DynY : " + str(dynY))
        distult = distance(botx, boty, dynX, dynY)
        if WriteFlag == 0:
            logger.info("[T:" + str(simtime) + "] " + "Dist to collision point: " + str(distult))
        rospy.loginfo("Dist to collision point: " + str(distult))
        if distult < safedist:
            rospy.loginfo("Collision!")
            if WriteFlag == 0:
                logger.info("[T:" + str(simtime) + "] " + "Collision!")
                os._exit(1)


def slope(x1, y1, x2, y2):  # Line slope given two points:
    return (y2 - y1) / (x2 - x1)


def angle(s1, s2):
    return math.degrees(math.atan((s2 - s1) / (1 + (s2 * s1))))


def angle_between(lineA, lineB):
    slope1 = slope(lineA[0][0], lineA[0][1], lineA[1][0], lineA[1][1])
    slope2 = slope(lineB[0][0], lineB[0][1], lineB[1][0], lineB[1][1])

    ang = angle(slope1, slope2)
    if ang < 0:
        ang = ang * (-1.0)

    return ang

def get_rotation (msg):
    global yawBot
    orientation_q = msg.pose.pose.orientation
    orientation_list = [orientation_q.x, orientation_q.y, orientation_q.z, orientation_q.w]
    (roll, pitch, yaw) = euler_from_quaternion (orientation_list)
    yaw_degrees = yaw * 180.0 /math.pi
    if yaw_degrees <0:
        yaw_degrees += 360.0
    #rospy.loginfo("Yaw:" + str(yaw_degrees))
    yawBot = yaw_degrees

def callback(msg):
    global BotDist
    global yawBot
    global ExitFlag
    get_rotation(msg)
    botx = msg.pose.pose.position.x
    boty = msg.pose.pose.position.y
    botx = botx + BotDist * math.cos(math.radians(yawBot))
    boty = boty + BotDist * math.sin(math.radians(yawBot))
    rospy.loginfo("BotX: " + str(botx))
    rospy.loginfo("BotY: " + str(boty))
    detect = Detection()
    detect.show_gazebo_models(botx, boty)
    gc.collect()




def main():
    rospy.init_node('location_monitor') #Node in gang bringen
    try:
        rospy.Subscriber("/odom", Odometry, callback) #Koordinaten des bots werden durch /odom geschickt und an callback gesendet
        gc.collect()
        rospy.spin()
    except rospy.ROSInterruptException:
        pass

if __name__ == '__main__':
    main()
